@@ -6,6 +6,7 @@ import (
 	"fmt"
 	"io"
 	"regexp"
+	"sync"
 
 	"github.com/shenwei356/bio/seq"
 	"github.com/shenwei356/xopen"
@@ -25,6 +26,24 @@ var ErrNoContent = errors.New("fastx: no content found")
 
 var bufSize = 65536
 
+var poolReader = &sync.Pool{New: func() interface{} {
+	t := &Reader{}
+
+	t.buf = make([]byte, bufSize)
+
+	t.buffer = bytes.NewBuffer(make([]byte, 0, defaultBytesBufferSize))
+	t.seqBuffer = bytes.NewBuffer(make([]byte, 0, defaultBytesBufferSize))
+	t.qualBuffer = bytes.NewBuffer(make([]byte, 0, defaultBytesBufferSize))
+
+	t.record = &Record{
+		ID:   nil,
+		Name: nil,
+		Desc: nil,
+		Seq:  &seq.Seq{}, // can't be nil
+	}
+	return t
+}}
+
 // Reader seamlessly parse both FASTA and FASTQ formats
 type Reader struct {
 	fh *xopen.Reader // file handle, xopen is such a wonderful package
@@ -50,7 +69,35 @@ type Reader struct {
 	qualBuffer      *bytes.Buffer
 	record          *Record
 
-	Err error // Current error
+	// only for compatibility of empty files
+	Err error
+}
+
+func (fastxReader *Reader) Reset() {
+	n := bufSize - len(fastxReader.buf)
+	for i := 0; i < n; i++ {
+		fastxReader.buf = append(fastxReader.buf, 0)
+	}
+	fastxReader.r = 0
+	fastxReader.buffer.Reset()
+	fastxReader.needMoreCheckOfBuf = false
+	fastxReader.lastByte = 0
+	fastxReader.checkSeqType = true
+	fastxReader.lastPart = false
+	fastxReader.finished = false
+
+	fastxReader.firstseq = true
+	fastxReader.delim = 0
+	fastxReader.IsFastq = false
+
+	fastxReader.seqBuffer.Reset()
+	fastxReader.qualBuffer.Reset()
+
+	fastxReader.head = nil
+	fastxReader.seq = nil
+	fastxReader.qual = nil
+
+	fastxReader.Err = nil
 }
 
 // regexp for checking idRegexp string.
@@ -72,13 +119,14 @@ var defaultBytesBufferSize = 10 << 20
 //
 // Parameters:
 //
-//        t            sequence alphabet
-//                     if nil is given, it will guess alphabet by the first record
-//        file         file name, "-" for stdin
-//        idRegexp     id parsing regular expression string, must contains "(" and ")" to capture matched ID
-//                     "" for default value: `^([^\s]+)\s?`
-//                     if record head does not match the idRegxp, whole name will be the id
+//	t            sequence alphabet
+//	             if nil is given, it will guess alphabet by the first record
+//	file         file name, "-" for stdin
+//	idRegexp     id parsing regular expression string, must contains "(" and ")" to capture matched ID
+//	             "" for default value: `^([^\s]+)\s?`
+//	             if record head does not match the idRegxp, whole name will be the id
 //
+// Please call reader.Close() afer using the records!!!
 func NewReader(t *seq.Alphabet, file string, idRegexp string) (*Reader, error) {
 	var r *regexp.Regexp
 	if idRegexp == "" {
@@ -101,35 +149,20 @@ func NewReader(t *seq.Alphabet, file string, idRegexp string) (*Reader, error) {
 	fh, err := xopen.Ropen(file)
 	if err != nil {
 		if err == xopen.ErrNoContent {
-			return &Reader{
-					fh:       nil,
-					finished: true,
-					lastPart: true,
-					Err:      io.EOF,
-				},
-				nil
+			fastxReader := poolReader.Get().(*Reader)
+			fastxReader.Reset()
+			fastxReader.Err = io.EOF // so the first call of Read will return an io.EOF error.
+			return fastxReader, nil
 		}
 		return nil, fmt.Errorf("fastx: %s", err)
 	}
 
-	fastxReader := &Reader{
-		fh:           fh,
-		buf:          make([]byte, bufSize),
-		t:            t,
-		IDRegexp:     r,
-		firstseq:     true,
-		checkSeqType: true,
-	}
-	fastxReader.buffer = bytes.NewBuffer(make([]byte, 0, defaultBytesBufferSize))
-	fastxReader.seqBuffer = bytes.NewBuffer(make([]byte, 0, defaultBytesBufferSize))
-	fastxReader.qualBuffer = bytes.NewBuffer(make([]byte, 0, defaultBytesBufferSize))
+	fastxReader := poolReader.Get().(*Reader)
+	fastxReader.Reset()
 
-	fastxReader.record = &Record{
-		ID:   nil,
-		Name: nil,
-		Desc: nil,
-		Seq:  &seq.Seq{},
-	}
+	fastxReader.fh = fh
+	fastxReader.t = t
+	fastxReader.IDRegexp = r
 
 	return fastxReader, nil
 }
@@ -138,13 +171,14 @@ func NewReader(t *seq.Alphabet, file string, idRegexp string) (*Reader, error) {
 //
 // Parameters:
 //
-//        t            sequence alphabet
-//                     if nil is given, it will guess alphabet by the first record
-//        file         an io.Reader
-//        idRegexp     id parsing regular expression string, must contains "(" and ")" to capture matched ID
-//                     "" for default value: `^([^\s]+)\s?`
-//                     if record head does not match the idRegxp, whole name will be the id
+//	t            sequence alphabet
+//	             if nil is given, it will guess alphabet by the first record
+//	file         an io.Reader
+//	idRegexp     id parsing regular expression string, must contains "(" and ")" to capture matched ID
+//	             "" for default value: `^([^\s]+)\s?`
+//	             if record head does not match the idRegxp, whole name will be the id
 //
+// Please call reader.Close() afer using the records!!!
 func NewReaderFromIO(t *seq.Alphabet, ioReader io.Reader, idRegexp string) (*Reader, error) {
 	var r *regexp.Regexp
 	if idRegexp == "" {
@@ -169,30 +203,24 @@ func NewReaderFromIO(t *seq.Alphabet, ioReader io.Reader, idRegexp string) (*Rea
 		panic(err)
 	}
 
-	fastxReader := &Reader{
-		fh:           fh,
-		buf:          make([]byte, bufSize),
-		t:            t,
-		IDRegexp:     r,
-		firstseq:     true,
-		checkSeqType: true,
-	}
-	fastxReader.buffer = bytes.NewBuffer(make([]byte, 0, 1<<20))
-	fastxReader.seqBuffer = bytes.NewBuffer(make([]byte, 0, 1<<20))
-	fastxReader.qualBuffer = bytes.NewBuffer(make([]byte, 0, 1<<20))
+	fastxReader := poolReader.Get().(*Reader)
+	fastxReader.Reset()
 
-	fastxReader.record = &Record{
-		ID:   nil,
-		Name: nil,
-		Desc: nil,
-		Seq:  &seq.Seq{},
-	}
+	fastxReader.fh = fh
+	fastxReader.t = t
+	fastxReader.IDRegexp = r
 
 	return fastxReader, nil
 }
 
-// Close closes the reader
+// Close cleans up everything, the most important thing is recyling the reader.
+// Please do remember to calls this method!!!
 func (fastxReader *Reader) Close() {
+	poolReader.Put(fastxReader)
+}
+
+// close closes the file handler
+func (fastxReader *Reader) close() {
 	if fastxReader.fh != nil {
 		fastxReader.fh.Close()
 	}
@@ -203,17 +231,13 @@ func (fastxReader *Reader) Close() {
 // the current record will change after your another call of this method.
 // So, you could use record.Clone() to make a copy.
 func (fastxReader *Reader) Read() (*Record, error) {
-	fastxReader.read()
-	return fastxReader.record, fastxReader.Err
-}
-
-func (fastxReader *Reader) read() {
 	if fastxReader.lastPart && fastxReader.finished {
-		fastxReader.Err = io.EOF
-		return
+		return nil, io.EOF
 	}
-	if fastxReader.Err != nil {
-		return
+
+	if fastxReader.Err != nil { // only for empty file
+		err := fastxReader.Err
+		return nil, err
 	}
 
 	var n int
@@ -227,14 +251,11 @@ func (fastxReader *Reader) read() {
 				if err == io.EOF {
 					fastxReader.lastPart = true
 				} else {
-					fastxReader.Err = err
-					fastxReader.Close()
-					return
+					return nil, err
 				}
 			} else if n == 0 {
-				fastxReader.Err = io.ErrUnexpectedEOF
-				fastxReader.Close()
-				return
+				fastxReader.close()
+				return nil, io.ErrUnexpectedEOF
 			}
 
 			// last part of file OR just because reader not fulfill the buf,
@@ -246,6 +267,7 @@ func (fastxReader *Reader) read() {
 			fastxReader.r = 0 /// TO CHECK
 		}
 
+		// check seq type via the first non-empty charator, run for onece
 		if fastxReader.checkSeqType {
 			pn := 0
 		FORCHECK:
@@ -266,17 +288,15 @@ func (fastxReader *Reader) read() {
 					pn++
 					if pn > 100 {
 						if i > 10240 { // ErrNotFASTXFormat
-							fastxReader.Err = ErrNotFASTXFormat
-							fastxReader.Close()
-							return
+							fastxReader.close()
+							return nil, ErrNotFASTXFormat
 						}
 					}
 					// break FORCHECK
 				default: // not typical FASTA/Q
 					// if i > 10240 || fastxReader.lastPart { // ErrNotFASTXFormat
-					fastxReader.Err = ErrNotFASTXFormat
-					fastxReader.Close()
-					return
+					fastxReader.close()
+					return nil, ErrNotFASTXFormat
 					// }
 				}
 			}
@@ -305,7 +325,7 @@ func (fastxReader *Reader) read() {
 						fastxReader.buffer.WriteByte('\n')
 					}
 
-					// we have to avoid the case of quality line starts with "@"
+					// we have to avoid the case of quality line starting with "@"
 					shorterQual, err = fastxReader.parseRecord()
 					if fastxReader.IsFastq && err != nil && err == ErrUnequalSeqAndQual {
 						if shorterQual {
@@ -315,13 +335,13 @@ func (fastxReader *Reader) read() {
 							fastxReader.r += i + 1
 							continue FORSEARCH
 						}
-						fastxReader.Err = ErrBadFASTQFormat
-						return
+						fastxReader.close()
+						return nil, ErrBadFASTQFormat
 					}
 					fastxReader.buffer.Reset()
 					fastxReader.needMoreCheckOfBuf = true
 					fastxReader.r += i + 1
-					return
+					return fastxReader.record, nil
 				}
 				// inline >/@
 				fastxReader.buffer.Write(fastxReader.buf[fastxReader.r : fastxReader.r+i+1])
@@ -334,14 +354,13 @@ func (fastxReader *Reader) read() {
 			if fastxReader.lastPart {
 				_, err = fastxReader.parseRecord()
 				if err != nil { // no any chance
-					fastxReader.Err = err
-					fastxReader.Close()
-					return
+					fastxReader.close()
+					return nil, err
 				}
 				fastxReader.buffer.Reset()
-				fastxReader.Close()
+				fastxReader.close()
 				fastxReader.finished = true
-				return
+				return fastxReader.record, nil
 			}
 			fastxReader.needMoreCheckOfBuf = false
 			break FORSEARCH
@@ -418,13 +437,10 @@ func (fastxReader *Reader) parseRecord() (bool, error) {
 	if len(fastxReader.head) == 0 && len(fastxReader.seq) == 0 {
 		return false, io.EOF
 	}
-	// new record
-	if fastxReader.IsFastq {
-		// fastxReader.record, fastxReader.Err = NewRecordWithQual(fastxReader.t,
-		// 	fastxReader.parseHeadID(fastxReader.head), fastxReader.head,
-		// 	fastxReader.parseHeadDesc(fastxReader.head),
-		// 	fastxReader.seq, fastxReader.qual)
 
+	var err error
+	// new record
+	if fastxReader.IsFastq { // fastq
 		fastxReader.record.Seq.Alphabet = fastxReader.t
 		fastxReader.record.ID, fastxReader.record.Desc = parseHeadIDAndDesc(fastxReader.IDRegexp, fastxReader.head)
 		fastxReader.record.Name = fastxReader.head
@@ -432,43 +448,26 @@ func (fastxReader *Reader) parseRecord() (bool, error) {
 		fastxReader.record.Seq.Qual = fastxReader.qual
 
 		if seq.ValidateSeq {
-			fastxReader.Err = fastxReader.t.IsValid(fastxReader.seq)
+			err = fastxReader.t.IsValid(fastxReader.seq)
 		}
 
 		if len(fastxReader.seq) != len(fastxReader.qual) {
-			fastxReader.Err = fmt.Errorf("seq: unmatched length of sequence (%d) and quality (%d)",
+			err = fmt.Errorf("seq: unmatched length of sequence (%d) and quality (%d)",
 				len(fastxReader.seq), len(fastxReader.qual))
 		}
 
-	} else {
-		// fastxReader.record, fastxReader.Err = NewRecord(fastxReader.t,
-		// 	fastxReader.parseHeadID(fastxReader.head), fastxReader.head,
-		// 	fastxReader.parseHeadDesc(fastxReader.head),
-		// 	fastxReader.seq)
-
+	} else { // fasta
 		fastxReader.record.Seq.Alphabet = fastxReader.t
 		fastxReader.record.ID, fastxReader.record.Desc = parseHeadIDAndDesc(fastxReader.IDRegexp, fastxReader.head)
 		fastxReader.record.Name = fastxReader.head
 		fastxReader.record.Seq.Seq = fastxReader.seq
 
 		if seq.ValidateSeq {
-			fastxReader.Err = fastxReader.t.IsValid(fastxReader.seq)
+			err = fastxReader.t.IsValid(fastxReader.seq)
 		}
 	}
 
-	if fastxReader.Err != nil {
-		fastxReader.Close()
-	}
-
-	return false, fastxReader.Err
-}
-
-func (fastxReader *Reader) parseHeadID(head []byte) []byte {
-	return parseHeadID(fastxReader.IDRegexp, head)
-}
-
-func (fastxReader *Reader) parseHeadDesc(head []byte) []byte {
-	return parseHeadDesc(fastxReader.IDRegexp, head)
+	return false, err
 }
 
 // ParseHeadID parse ID from head by IDRegexp. not used.
@@ -494,6 +493,9 @@ func parseHeadIDAndDesc(idRegexp *regexp.Regexp, head []byte) ([]byte, []byte) {
 					break
 				}
 			}
+			if j >= e {
+				return head[0:i], emptyByteSlice
+			}
 			return head[0:i], head[j:]
 		}
 		if i := bytes.IndexByte(head, '\t'); i > 0 {
@@ -506,6 +508,9 @@ func parseHeadIDAndDesc(idRegexp *regexp.Regexp, head []byte) ([]byte, []byte) {
 					break
 				}
 			}
+			if j >= e {
+				return head[0:i], emptyByteSlice
+			}
 			return head[0:i], head[j:]
 		}
 		return head, emptyByteSlice
@@ -518,57 +523,6 @@ func parseHeadIDAndDesc(idRegexp *regexp.Regexp, head []byte) ([]byte, []byte) {
 	return found[1], emptyByteSlice
 }
 
-// parseHeadID parse ID from head by IDRegexp
-func parseHeadID(idRegexp *regexp.Regexp, head []byte) []byte {
-	if isUsingDefaultIDRegexp {
-		if i := bytes.IndexByte(head, ' '); i > 0 {
-			return head[0:i]
-		}
-		if i := bytes.IndexByte(head, '\t'); i > 0 {
-			return head[0:i]
-		}
-		return head
-	}
-
-	found := idRegexp.FindSubmatch(head)
-	if found == nil { // not match
-		return head
-	}
-	return found[1]
-}
-
-// parseHeadDesc returns description of header
-func parseHeadDesc(idRegexp *regexp.Regexp, head []byte) []byte {
-	if isUsingDefaultIDRegexp {
-		if i := bytes.IndexByte(head, ' '); i > 0 {
-			// return bytes.TrimLeft(head[i:], " \t")
-			e := len(head)
-			for i++; i < e; i++ {
-				if head[i] == ' ' || head[i] == '\t' {
-					i++
-				} else {
-					break
-				}
-			}
-			return head[i:]
-		}
-		if i := bytes.IndexByte(head, '\t'); i > 0 {
-			// return bytes.TrimLeft(head[i:], " \t")
-			e := len(head)
-			for i++; i < e; i++ {
-				if head[i] == ' ' || head[i] == '\t' {
-					i++
-				} else {
-					break
-				}
-			}
-			return head[i:]
-		}
-		return []byte{}
-	}
-	return []byte{}
-}
-
 // Alphabet returns Alphabet of the file
 func (fastxReader *Reader) Alphabet() *seq.Alphabet {
 	if fastxReader.t == nil {
